
#define ADC_OUT  0          // channel for AC output sample

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define NPWM  400         // 200 for this and 800 for PPWM for 50Hz and 20kHz PWM. Both must be integers.
#define PPWM  800         

      
uint8_t uf,oen,v1low, lv_stopped,int_f;
uint16_t pcount,icount;
int32_t vpwr;
int sst;
int l[NPWM+1];
float pwr;
uint8_t dbounce;
double Input, Output, Setpoint,errSum, lastErr,error,dErr, kp, ki, kd;
float ac_output, ac_setpoint;

void init_vars(void);

uint8_t init_done; 
uint8_t tt=0;

void setup()
  {
  int i;
  float t,u;

  for(t=0.0,i=0; i <= (NPWM-1); i++,t += 2.0*3.14159/ (float) (NPWM))
    {   
    u = 20000.0 * sin(t);   
    l[i] =  (int)u;  
    
    }
 
  pinMode(9, OUTPUT);     // V1 drive
  pinMode(10, OUTPUT);    // V2 drive
  pinMode(12, OUTPUT);    // inverter ON/OFF LED, pin 15 on nano, Arduino D12
  pinMode(8,INPUT);       // continuous on/off pin, driven by nano2
  pinMode(5,OUTPUT);      // used for IR2110 shutdown drive, LOW = shutdown
  pinMode(7,OUTPUT);      // DSO sync pulse, each 50Hz
  noInterrupts();

  TCCR1A = _BV(COM1A1) |  _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = OCR1B = PPWM/2;
  ICR1 = PPWM*2;
  TIMSK1 |= (1 << TOIE1);

  cbi(PORTD,5);   // disable gate drive output. data direction flag set by above statement.
  oen = 0;        // output enable. = 1 to run inverter, = 0 to stop inverter
  init_vars();
  init_done = 1;
  
  interrupts();             // enable all interrupts
  TIMSK0 = 0;               // except arduino clock

  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;  

  }



//
// 20Khz SPWM, code has only 50uS to run, takes about 10uS
//
ISR(TIMER1_OVF_vect)        
 {   
  long c;
  c = (l[pcount] * vpwr) >> 16;   
  OCR1B = PPWM/2 + c;
  OCR1A = PPWM/2 - c; 
  
  pcount++;
  icount++;
  if(icount == NPWM/2)          // pcount  will = NPWM/2, at each 1/2 wave, at the start
    {
    icount = 0;  
    uf=1;                     // enable one PID control loop execution, in loop()
    if (v1low == 1)           // if first 1/2 wave..
      {
       sbi(PORTD,7);          //DSO pulse to tell mains sync pulse occurred in maybe wrong time
       v1low = 0;             // toggle to 2nd half wave
      }                         
    else
      {
       cbi(PORTD,7); 
       v1low = 1;             // toggle to 1st half wave
      }
    }
    
  if (pcount >= NPWM)
    pcount = 0;
}


void init_vars()      // called prior to setting oen = 1, as well as from setup()
  {
  ICR1 = PPWM;
  lv_stopped = 0;  
  sst = 0;      // slow start counter 0 - 251
  v1low = 1;    // half wave output flag, 0 or 1
  icount = 0;
  pcount = 0;   // counter for sine wave lookup table
  uf = 0;       // set to 1 to enable AC output voltage feedback code to execute
  pwr = 0.0;    // PWM duty cycle analog, 0.0 to 1.0
  vpwr= 0;      // integer equivalent of pwr, 0 to PPWM
  dbounce = 0;  // debounce counter for input switch 
  kp = 0.1;     // set up PID for AC output voltage control
  ki = 0.0;
  kd = 0.01;
  errSum = 0.0;        // initialise PID
  lastErr = 0.0;
  cbi(PORTD,4);  
  }


void do_pid()
{   
   double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}


void loop()
{
  float ch0;

  // sample channel..
  // In the past, I low pass filtered this. not now, no need thanks to active LP filter on Vfb
  ac_output = (float)analogRead(ADC_OUT) / 1024.0; // scale 0 - 5V to 0.0 to 1.0
 
  if(uf == 1  )    // this will execute at 100Hz
      {            // it runs at near zero volt crossing of AC output       
      uf=0; 
      ac_setpoint = 0.55;       // for same Vfb as EG8010. change AC output voltage via Vfb voltage divider trimpot
      if (oen == 1) sst++;      // slow start. If output is enabled, keep starting up. sst = output pwm% or approx. voltage.
      if (sst > 251)sst = 251;  // stop increasing when sst has got to the end of slow start. when = 251, this enables PID. See below..
      if (oen == 0) sst--;      // if stopping, slow stop
      if (sst <= 0)             // once fully stopped, no more changes to sst needed
        {
        sst = 0;
        cbi(PORTD,5);     // pull IR2184 shutdown LOW to disable gate drive output
        }
      else
        sbi(PORTD,5);     // pull it HIGH, to enable output

      if(sst > 0 && sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {                          // sst can increase or decrease, giving the slow start and slow stop.
        if (oen == 1)              // slow stop? I think of it as de-gauss
          { 
          if(ac_output < ac_setpoint)
            pwr += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID.
          else                    // During slow start need to ensure output is tracking setpoint
            pwr -= 1.0/250.0;     // should output get close enough to matter.
          }
        else
          pwr -= 1.0 /250.0;      // this will execute in slow stop
        }
        
      if (sst == 251)       // once sst = 251, run PID. the above bang-bang control will NOT have executed, because sst = 251
        {                   
        Setpoint = ac_setpoint;     
        Input = ac_output;  
        do_pid();
        pwr = pwr + Output;
        }
        
      if (pwr > 0.99) pwr = 0.99;     // clamp. Do math in floating point to prevent any integer overflow
      if (pwr < 0.01) pwr = 0.01;     // it's slower. We could do clamp after conversion to int but it's your mosfets...
      vpwr = (int)((float)PPWM * pwr);// finally apply required PWM duty power factor to vpwr     
      check_switch_cont(); 
      }

}

void check_switch_cont()
{   // called at 100 Hz
  if ((PINB & 0x01) == 0x01)     
    {
    if (init_done == 0 && sst == 0)
      {
      init_vars();
      init_done = 1;
      }  
    oen = 1;      // slow start
    sbi(PORTB,4); // and light LED on D12, pin 15 of nano
    }

  if (((PINB & 0x01) == 0) )       
    {
    init_done = 0;
    oen = 0;    // slow stop
    cbi(PORTB,4); // and blank LED on D12, pin 15 of nano
    }         
}
